#include "IDataCollector.h"
#include "nsIVariant.h"
#include <nsCOMPtr.h>
#include <nsCOMArray.h>
#include <nsComponentManagerUtils.h>
#include <nsStringAPI.h>
#include "ICIS.h"
#include <stdlib.h>
#include <nsGenericFactory.h>
// For JSDOUBLE_IS_FINITE... (XXX)
#include <jspubtd.h>
#include <jsnum.h>

class CachedDataToCanvasRequest
  : public nsISupports
{
public:
  CachedDataToCanvasRequest(IDataToCanvasRequest* aDTCR)
  {
    nsresult ret;
    ret = aDTCR->GetGlyph(&mGlyph);
    NS_ENSURE_SUCCESS(ret,);
    ret = aDTCR->GetFirstPoint(&mFirstPoint);
    NS_ENSURE_SUCCESS(ret,);
    ret = aDTCR->GetLastPoint(&mLastPoint);
    NS_ENSURE_SUCCESS(ret,);

    nsString drawColour;
    ret = aDTCR->GetDrawColour(drawColour);
    NS_ENSURE_SUCCESS(ret,);
    mDrawColour =  do_CreateInstance(NS_VARIANT_CONTRACTID, &ret);
    NS_ENSURE_SUCCESS(ret,);
    mDrawColour->SetAsDOMString(drawColour);
    PRUint16 dt;
    mDrawColour->GetDataType(&dt);
    NS_ASSERTION(dt == nsIDataType::VTYPE_DOMSTRING,
                 "Buggy nsIVariant set wrong type");

    ret = aDTCR->GetXVariableMin(&mxVariableMin);
    NS_ENSURE_SUCCESS(ret,);
    ret = aDTCR->GetXVariableMax(&mxVariableMax);
    NS_ENSURE_SUCCESS(ret,);
    ret = aDTCR->GetYVariableMin(&myVariableMin);
    NS_ENSURE_SUCCESS(ret,);
    ret = aDTCR->GetYVariableMax(&myVariableMax);
    NS_ENSURE_SUCCESS(ret,);
    ret = aDTCR->GetXMultiplier(&mxMultiplier);
    NS_ENSURE_SUCCESS(ret,);
    ret = aDTCR->GetYMultiplier(&myMultiplier);
    NS_ENSURE_SUCCESS(ret,);
    ret = aDTCR->GetXOffset(&mxOffset);
    NS_ENSURE_SUCCESS(ret,);
    ret = aDTCR->GetYOffset(&myOffset);
    NS_ENSURE_SUCCESS(ret,);
    ret = aDTCR->GetXVariable(&mxVariable);
    NS_ENSURE_SUCCESS(ret,);
    ret = aDTCR->GetYVariable(&myVariable);
    NS_ENSURE_SUCCESS(ret,);

    ret = aDTCR->GetRenderContext(getter_AddRefs(mRenderContext));
    NS_ENSURE_SUCCESS(ret,);
    mDTCR = aDTCR;
  }

  NS_DECL_ISUPPORTS;

  nsCOMPtr<IDataToCanvasRequest> mDTCR;
  nsCOMPtr<nsIDOMCanvasRenderingContext2D> mRenderContext;
  nsCOMPtr<nsIWritableVariant> mDrawColour;
  PRUint32 mGlyph, mFirstPoint, mLastPoint, mxVariable, myVariable;
  PRFloat64 mxVariableMin, mxVariableMax, myVariableMin, myVariableMax;
  PRFloat64 mxMultiplier, myMultiplier, mxOffset, myOffset;

  void Draw(double* data, PRUint32 start, PRUint32 end, PRUint32 gap);
};

NS_IMPL_ISUPPORTS0(CachedDataToCanvasRequest)

class DataCollector
  : public IDataCollector,
    public cellml_services_IIntegrationProgressObserver
{
public:
  DataCollector();
  ~DataCollector();

  NS_DECL_ISUPPORTS;
  NS_DECL_IDATACOLLECTOR;
  NS_DECL_CELLML_SERVICES_IINTEGRATIONPROGRESSOBSERVER;

private:
  PRUint32 mStatus;
  PRUint32 mnConstants, mnVariables;
  double* mNotifiedMinimums, * mNotifiedMaximums, * mTrueMinimums,
        * mTrueMaximums;
  PRUint32 mDataCapacity, mDataSize;
  double* mData;
  nsCOMArray<IDataCollectorObserver> mObservers;
  nsCOMArray<CachedDataToCanvasRequest> mDTCRs;
  PRFloat64 mBoundStart;

  void NotifyVariableChanged(PRUint32 var);
};

NS_IMPL_ISUPPORTS2(DataCollector, IDataCollector,
                   cellml_services_IIntegrationProgressObserver);

DataCollector::DataCollector()
  : mStatus(IDataCollector::STATUS_NOTREADY),
    mNotifiedMinimums(NULL), mNotifiedMaximums(NULL), mTrueMinimums(NULL),
    mTrueMaximums(NULL), mDataSize(0), mData(NULL)
{
}

DataCollector::~DataCollector()
{
  if (mData != NULL)
    free(mData);
  if (mNotifiedMinimums != NULL)
    delete [] mNotifiedMinimums;
  if (mNotifiedMaximums != NULL)
    delete [] mNotifiedMaximums;
  if (mTrueMinimums != NULL)
    delete [] mTrueMinimums;
  if (mTrueMaximums != NULL)
    delete [] mTrueMaximums;
}

NS_IMETHODIMP
DataCollector::GetStatus(PRUint32* aStatus)
{
  *aStatus = mStatus;
  return NS_OK;
}

NS_IMETHODIMP
DataCollector::Initialise(PRUint32 nConstants, PRUint32 nVariables,
                          PRFloat64 aBoundStart, PRFloat64 aMaxDensity)
{
  if (mStatus != IDataCollector::STATUS_NOTREADY)
    return NS_ERROR_ALREADY_INITIALIZED;

  mnConstants = nConstants;
  mnVariables = nVariables + 1;

  // Next, allocate the datastructures...
  mNotifiedMinimums = new double[mnVariables];
  mNotifiedMaximums = new double[mnVariables];
  mTrueMinimums = new double[mnVariables];
  mTrueMaximums = new double[mnVariables];
  mDataCapacity = 10;
  mData = (double*)malloc(sizeof(double) * mDataCapacity * mnVariables);
  mStatus = IDataCollector::STATUS_INPROGRESS;

  mBoundStart = aBoundStart;
  return NS_OK;
}

NS_IMETHODIMP
DataCollector::GetNotifiedMinimum(PRUint32 aVariable, double* aMinimum)
{
  if (aVariable >= mnVariables)
    return NS_ERROR_INVALID_ARG;
  if (mDataSize == 0)
  {
    *aMinimum = -1.0;
    return NS_OK;
  }
  *aMinimum = mNotifiedMinimums[aVariable];
  return NS_OK;
}

NS_IMETHODIMP
DataCollector::GetNotifiedMaximum(PRUint32 aVariable, double* aMaximum)
{
  if (aVariable >= mnVariables)
    return NS_ERROR_INVALID_ARG;
  if (mDataSize == 0)
  {
    *aMaximum = 1.0;
    return NS_OK;
  }
  *aMaximum = mNotifiedMaximums[aVariable];
  return NS_OK;
}

NS_IMETHODIMP
DataCollector::GetRawData(PRUint32 aVariable, PRUint32 aPoint, double* aData)
{
  if (aVariable >= mnVariables)
    return NS_ERROR_INVALID_ARG;
  if (aPoint >= mDataSize)
    return NS_ERROR_INVALID_ARG;
  *aData = mData[aVariable + mnVariables * aPoint];
  return NS_OK;
}

NS_IMETHODIMP
DataCollector::AddDataCollectorObserver(IDataCollectorObserver* aNewObserver)
{
  mObservers.AppendObject(aNewObserver);
  return NS_OK;
}

NS_IMETHODIMP
DataCollector::RemoveDataCollectorObserver(IDataCollectorObserver* aObserver)
{
  mObservers.RemoveObject(aObserver);
  return NS_OK;
}

NS_IMETHODIMP
DataCollector::AddDataToCanvasRequest(IDataToCanvasRequest* aDTCR)
{
  CachedDataToCanvasRequest* cdtcr = new CachedDataToCanvasRequest(aDTCR);
  mDTCRs.AppendObject(cdtcr);
  
  // If theres any suitable data available, draw it straight away...
  if (mDataSize > 0)
  {
    if (cdtcr->mFirstPoint < mDataSize &&
        cdtcr->mFirstPoint <= cdtcr->mLastPoint)
    {
      PRUint32 lp = mDataSize - 1;
      if (cdtcr->mLastPoint < lp)
        lp = cdtcr->mLastPoint;
      cdtcr->Draw(mData, cdtcr->mFirstPoint, lp, mnVariables);
    }
  }

  return NS_OK;
}

NS_IMETHODIMP
DataCollector::RemoveDataToCanvasRequest(IDataToCanvasRequest* aDTCR)
{
  PRUint32 i, l = mDTCRs.Count();
  for (i = 0; i < l; i++)
    if (mDTCRs[i]->mDTCR == aDTCR)
    {
      mDTCRs.RemoveObjectAt(i);
      return NS_OK;
    }
  return NS_OK;
}

NS_IMETHODIMP
DataCollector::ComputedConstants(PRFloat64 *values, PRUint32 sizeof_values)
{
  // TODO: Store, and add a way to fetch these.
  return NS_OK;
}

NS_IMETHODIMP
DataCollector::Results(PRFloat64 *state, PRUint32 sizeof_state)
{
  if (sizeof_state == 0)
    return NS_OK;
  PRUint32 pcount = sizeof_state / mnVariables;

  if (pcount * mnVariables != sizeof_state)
    return NS_ERROR_INVALID_ARG;

  if (mDataCapacity < mDataSize + pcount)
  {
    mDataCapacity = (mDataSize + pcount) * 3 / 2 + 10;
    mData = (double*)realloc(mData,
                             sizeof(double) * mDataCapacity * mnVariables);
  }

  PRBool* variablesToChange = new PRBool[mnVariables];
  memset(variablesToChange, 0, sizeof(PRBool) * mnVariables);

  PRUint32 i, v = 0, x = mDataSize * mnVariables;
  for (i = 0; i <= sizeof_state; i++, v++, x++)
  {
    if (v >= mnVariables)
    {
      v = 0;
      mDataSize++;
      if (i == sizeof_state)
        break;
      mData[x] = state[i];
    }
    else
    {
      mData[x] = state[i];
    }

    PRBool newrange = PR_FALSE;

    PRFloat64 oldrange = mTrueMaximums[v] - mTrueMinimums[v];

    if (mDataSize == 0)
    {
      newrange = PR_TRUE;
      mTrueMinimums[v] = mData[x];
      mTrueMaximums[v] = mData[x];
    }
    else
    {
      if (mTrueMinimums[v] > mData[x])
      {
        newrange = PR_TRUE;
        mTrueMinimums[v] = mData[x];
      }
      else if (mTrueMaximums[v] < mData[x])
      {
        newrange = PR_TRUE;
        mTrueMaximums[v] = mData[x];
      }
    }
    if (newrange)
    {
      PRFloat64 range = mTrueMaximums[v] - mTrueMinimums[v];
      if (range == 0.0)
      {
        // First point, so just +/- 20% it (unless its 0)...
        if (mDataSize == 0)
        {
          if (mData[x] == 0.0)
          {
            mNotifiedMinimums[v] = -1.0;
            mNotifiedMaximums[v] = 1.0;
          }
          else
          {
            mNotifiedMinimums[v] = mData[x] * 0.9;
            mNotifiedMaximums[v] = mData[x] * 1.1;
            // In case its negative...
            if (mNotifiedMaximums[v] < mNotifiedMinimums[v])
            {
              PRFloat64 t = mNotifiedMaximums[v];
              mNotifiedMaximums[v] = mNotifiedMinimums[v];
              mNotifiedMinimums[v] = t;
            }
          }
          // Now notify the new value...
          variablesToChange[v] = PR_TRUE;
        }
        // if not the first point, range is still zero, don't notify...
      }
      else
      {
        if (oldrange == 0.0)
        {
          // The old true range is 0.0, which means that we used the 20% margin
          // or +/- 1.0 rule. We now set the new range with 10% expansion. This
          // may result in the notified range shrinking.
          PRFloat64 margin = range * 0.1;
          mNotifiedMinimums[v] = mTrueMinimums[v] - margin;
          mNotifiedMaximums[v] = mTrueMaximums[v] + margin;
          variablesToChange[v] = PR_TRUE;
        }
        else
        {
          // See if the current true range fits in the notified range...
          if (mTrueMinimums[v] < mNotifiedMinimums[v] ||
              mTrueMaximums[v] > mNotifiedMaximums[v])
          {
            PRFloat64 margin = range * 0.1;
            mNotifiedMinimums[v] = mTrueMinimums[v] - margin;
            mNotifiedMaximums[v] = mTrueMaximums[v] + margin;
            variablesToChange[v] = PR_TRUE;
          }
        }
      }
    }
  }

  // Figure out which variables need to be notified...
  for (i = 0; i < mnVariables; i++)
    if (variablesToChange[i])
      NotifyVariableChanged(i);

  // Having updated the scales, we now need to go through each
  // CachedDataToCanvasRequest and see if we can draw more points...
  PRUint32 start = mDataSize - pcount;
  for (i = 0; i < (PRUint32)mDTCRs.Count(); i++)
  {
    PRUint32 fp = start, lp = mDataSize - 1;
    if (mDTCRs[i]->mFirstPoint > fp)
      fp = mDTCRs[i]->mFirstPoint;
    if (mDTCRs[i]->mLastPoint < lp)
      lp = mDTCRs[i]->mLastPoint;

    if (fp > lp || fp >= mDataSize || lp < start)
      continue;

    mDTCRs[i]->Draw(mData, fp, lp, mnVariables);
  }

  delete [] variablesToChange;

  return NS_OK;
}

NS_IMETHODIMP
DataCollector::Done()
{
  mStatus = IDataCollector::STATUS_DONE;

  // Inform all our observers...
  PRUint32 i;
  for (i = 0; i < (PRUint32)mObservers.Count(); i++)
    mObservers[i]->Done();

  return NS_OK;
}

NS_IMETHODIMP
DataCollector::Failed(const char *errorMessage)
{
  mStatus = IDataCollector::STATUS_FAILED;

  // Inform all our observers...
  PRUint32 i;
  nsCString em(errorMessage);
  for (i = 0; i < (PRUint32)mObservers.Count(); i++)
    mObservers[i]->Failed(em);

  return NS_OK;
}

void
DataCollector::NotifyVariableChanged(PRUint32 var)
{
  PRUint32 i;
  for (i = 0; i < (PRUint32)mObservers.Count(); i++)
    mObservers[i]->RangeChanged(var, mNotifiedMinimums[var],
                                mNotifiedMaximums[var]);
}

void
CachedDataToCanvasRequest::Draw
(
 double* data,
 PRUint32 start,
 PRUint32 end,
 PRUint32 gap
)
{
  NS_ENSURE_TRUE(mRenderContext,);
  NS_ENSURE_TRUE((mxVariable < gap) && (myVariable < gap),);

  if (mGlyph == IDataToCanvasRequest::GLYPH_DOTS)
    mRenderContext->SetFillStyle(mDrawColour);
  else
    mRenderContext->SetStrokeStyle(mDrawColour);

  if (mGlyph == IDataToCanvasRequest::GLYPH_LINES)
  {
    // We need to include the previous point if available...
    if (start && (start - 1 >= mFirstPoint))
      start--;
    PRFloat64 *p = data + (start * gap);
    PRFloat64 *q = p + gap;
    PRUint32 i;
    if (end == start)
      return;
    for (i = start; i < end; i++, p += gap, q += gap)
    {
      PRFloat64 x1 = p[mxVariable], y1 = p[myVariable];
      PRFloat64 x2 = q[mxVariable], y2 = q[myVariable];

      PRUint32 validPointCount = 0;
      PRFloat64 xI[2], yI[2];
      
      if ((x1 == x2) && (y1 == y2))
        continue;

#define NEW_POINT(xval,yval) \
  if (validPointCount == 0 || xI[0] != xval || \
      yI[0] != yval) \
  { \
    xI[validPointCount] = xval; \
    yI[validPointCount] = yval; \
    validPointCount++; \
  }
      // Note: This only works for points known to be on the line.
#define POINT_ON_SEGMENT(x, y) ((((x1 <= x) && (x <= x2)) || \
                                 ((x2 <= x) && (x <= x1))) && \
                                (((y1 <= y) && (y <= y2)) || \
                                 ((y2 <= y) && (y <= y1))))

      if (((mxVariableMin <= x1) && (x1 <= mxVariableMax)) &&
          ((myVariableMin <= y1) && (y1 <= myVariableMax)))
      {
        // The first point is in the region...
        NEW_POINT(x1, y1);
      }

      if (((mxVariableMin <= x2) && (x2 <= mxVariableMax)) &&
          ((myVariableMin <= y2) && (y2 <= myVariableMax)))
      {
        // The second point is in the region...
        NEW_POINT(x2, y2);
      }
      
      if (validPointCount < 2)
      {
        PRFloat64 slope = (y2 - y1) / (x2 - x1);
        PRFloat64 intercept = y1 - slope * x1;
        if (JSDOUBLE_IS_FINITE(slope))
        {
          PRFloat64 yAtXmin = slope * mxVariableMin + intercept;
          if ((myVariableMin <= yAtXmin) && (yAtXmin <= myVariableMax) &&
              POINT_ON_SEGMENT(mxVariableMin, yAtXmin))
          {
            NEW_POINT(mxVariableMin, yAtXmin);
          }

          if (validPointCount < 2)
          {
            PRFloat64 yAtXmax = slope * mxVariableMax + intercept;
            if ((myVariableMin <= yAtXmax) && (yAtXmax <= myVariableMax) &&
                POINT_ON_SEGMENT(mxVariableMax, yAtXmax))
            {
              NEW_POINT(mxVariableMax, yAtXmax);
            }
          }
        }
        if (validPointCount < 2)
        {
          slope = (x2 - x1) / (y2 - y1);
          intercept = x1 - slope * y1;
          if (JSDOUBLE_IS_FINITE(slope))
          {
            PRFloat64 xAtYmin = slope * myVariableMin + intercept;
            if ((mxVariableMin <= xAtYmin) && (xAtYmin <= mxVariableMax) &&
                POINT_ON_SEGMENT(xAtYmin, myVariableMin))
            {
              NEW_POINT(xAtYmin, myVariableMin);
            }
            if (validPointCount < 2)
            {
              PRFloat64 xAtYmax = slope * myVariableMax + intercept;
              if ((mxVariableMin <= xAtYmax) && (xAtYmax <= mxVariableMax) &&
                  POINT_ON_SEGMENT(xAtYmax, myVariableMax))
              {
                NEW_POINT(xAtYmax, myVariableMax);
              }
            }
          }
        }
      }

      if (validPointCount != 2)
        continue;

      // We now have two points in the boundary. Stroke a path...
      mRenderContext->BeginPath();
      mRenderContext->MoveTo(mxMultiplier * xI[0] + mxOffset,
                             myMultiplier * yI[0] + myOffset);
      mRenderContext->LineTo(mxMultiplier * xI[1] + mxOffset,
                             myMultiplier * yI[1] + myOffset);
      mRenderContext->Stroke();
    }
  }
  else
  {
    PRFloat64 *p = data + (start * gap);
    PRUint32 i;
    for (i = start; i <= end; i++, p += gap)
    {
      PRFloat64 x = p[mxVariable], y = p[myVariable];
      if (x < mxVariableMin || x > mxVariableMax ||
          y < myVariableMin || y > myVariableMax)
        continue;
      x = mxMultiplier * x + mxOffset;
      y = myMultiplier * y + myOffset;
      switch (mGlyph)
      {
      case IDataToCanvasRequest::GLYPH_SQUARES:
        mRenderContext->StrokeRect(x - 5.0, y - 5.0, 9.0, 9.0);
        break;
      case IDataToCanvasRequest::GLYPH_CIRCLES:
        mRenderContext->BeginPath();
        mRenderContext->MoveTo(x + 4.0, y);
        mRenderContext->Arc(x, y, 4.0, 0.0, 6.28318530717959, PR_TRUE);
        mRenderContext->Stroke();
        break;
      case IDataToCanvasRequest::GLYPH_DIAMONDS:
        mRenderContext->Save();
        // Rotate by 45deg (1/4 pi radians)...
        mRenderContext->Translate(x, y);
        mRenderContext->Rotate(0.785398163397448);
        mRenderContext->StrokeRect(-5.0, -5.0, 9.0, 9.0);
        mRenderContext->Restore();
        break;
      case IDataToCanvasRequest::GLYPH_DOTS:
        mRenderContext->FillRect(x - 1.0, y - 1.0, 3.0, 3.0);
        break;
      }
    }
  }
}

NS_GENERIC_FACTORY_CONSTRUCTOR(DataCollector);

#define DATACOLLECTOR_CID \
{/*e97f3418-03fe-4cbf-9e66-d4cd328525d7*/\
 0xe97f3418, 0x03fe, 0x4cbf, {0xd4,0xcd,0x32,0x85,0x25,0xd7}}

static const nsModuleComponentInfo components[] = 
{
  {
    "DataCollector",
    DATACOLLECTOR_CID,
    "@cellml.org/tools/pcenv/datacollector;1",
    DataCollectorConstructor
  }
};
NS_IMPL_NSGETMODULE(DataCollector, components)
